@@ -1,6 +1,6 @@
 # High Availability package for DBI
 #
-# Copyright (c) 2003-2006 Henri Asseily <henri@asseily.com>. All rights reserved.
+# Copyright (c) 2003-2004 Henri Asseily <henri@asseily.com>. All rights reserved.
 # This program is free software; you can redistribute it and/or
 # modify it under the same terms as Perl itself.
 
@@ -11,7 +11,7 @@ use 5.006000;
 
 use constant DBIx_HA_DEBUG => 0;
 use Data::Dumper;
-use DBI 1.49 ();
+use DBI 1.44 ();
 use Sys::SigAction qw( set_sig_handler );
 use Exporter ();
 use strict;
@@ -20,20 +20,13 @@ use vars qw ( @ISA $prefix );
 
 our $loaded_Apache = 0;
 our $loaded_Apache_DBI = 0;
-our $logdir;
 
 BEGIN {
-	$DBIx::HA::VERSION = 0.99;
-
-	# Sample Fail Test Functions for different Database Servers
-	# They are used in the configuration 
-	# Input is ($ErrorID, $ErrorString)
-	# Output is boolean: If true, then we'll consider the error a critical condition, ok to failover
-	# If false, then DBIx::HA will not act on it and pass it straight through to the client
-	sub FTF_SybaseASE { $_[0] > 10 ? 0 : 1 }; # Fail Test Function for Sybase ASE
+	$DBIx::HA::VERSION = 0.91;
 }
 
 our $prefix = "[$$] DBIx::HA:           "; 
+my $logdir;
 
 sub initialize {
 	if ($Apache::VERSION) {
@@ -63,13 +56,6 @@ sub initialize {
 		if (! $DATABASE::conf{$dbname}->{'executetimeout'}) {
 			$DATABASE::conf{$dbname}->{'executetimeout'} = 10;
 		}
-		# add default failover rule based on DBI error
-		# default is to not request failover, whatever the error
-		# do not ever set the default to always request failover, as any SQL error would trigger failover!
-		if (! $DATABASE::conf{$dbname}->{'failtest_function'}) {
-			$DATABASE::conf{$dbname}->{'failtest_function'} = sub { 0 };
-		}
-		$DATABASE::conf{$dbname}->{'end_of_stack'} = 0;	# error condition reaching end of stack
 		foreach (@{$DATABASE::conf{$dbname}->{'db_stack'}}) {
 			# create an easy reverse-lookup table for finding the db server name from the dsn
 			$DBIx::HA::finddbserver{$_->[0]}  = $dbname;
@@ -108,7 +94,7 @@ sub _init_child {
 	foreach $dbname (keys %DATABASE::conf) {
 		# under application failover, maybe we already have an active db.
 		# set the active database to be the first in the stack unless we got it earlier.
-		_readsharedfile($dbname);
+		_readsharedfile($dbname) unless ($loaded_Apache);
 		_writesharedfile($dbname, 0) unless ($DATABASE::conf{$dbname}->{'active_db'});
 
 		# allow for connect on initialization
@@ -125,27 +111,23 @@ sub _readsharedfile {
 	my $dbname = shift;
 	if ($DATABASE::conf{$dbname}->{'failoverlevel'} eq 'application') {
 		# do this only if we're doing application failover and not process failover
-		if ($loaded_Apache && (-f "$logdir/DBIxHA_activedb_$dbname")) {
-			open IN, "$logdir/DBIxHA_activedb_$dbname";
-			my $dbidx = <IN>;
-			chomp $dbidx;
-			close IN;
-			if ($dbidx == -1) {	# we're told that we're at the end of the stack. No db server is available.
-				$DATABASE::conf{$dbname}->{'end_of_stack'} = 1;
-				return 0;
-			}
-			$DATABASE::conf{$dbname}->{'end_of_stack'} = 0;
-			if (($dbidx =~ /^\d+$/o) && $DATABASE::conf{$dbname}->{'db_stack'}->[$dbidx]) {
-				$DATABASE::conf{$dbname}->{'active_db'} = $DATABASE::conf{$dbname}->{'db_stack'}->[$dbidx];
-				$DBIx::HA::activeserver{$dbname}  = $dbidx;
-				unless ($Apache::ServerStarting == 1) {
-					my $r = Apache->request;
-					$r->notes("activedb_$dbname", $dbidx) if (ref $r);
+		if ($loaded_Apache) {
+			# $logdir = Apache::server_root_relative(undef,'logs'); # unnecessary since set during _writesharedfile on init
+			my $r = Apache->request;
+			if ($r && (! defined($r->notes("activedb_$dbname"))) && (-f "$logdir/DBIxHA_activedb_$dbname")) {
+				open IN, "$logdir/DBIxHA_activedb_$dbname";
+				my $dbidx = <IN>;
+				chomp $dbidx;
+				close IN;
+				if (($dbidx =~ /^\d+$/o) && $DATABASE::conf{$dbname}->{'db_stack'}->[$dbidx]) {
+					$DATABASE::conf{$dbname}->{'active_db'} = $DATABASE::conf{$dbname}->{'db_stack'}->[$dbidx];
+					$DBIx::HA::activeserver{$dbname}  = $dbidx;
+					$r->notes("activedb_$dbname", $dbidx);
+				} else {
+					warn "$prefix in _isactivedb: $dbname shared file has erroneous content, overwriting.\n";
+					_writesharedfile($dbname, $DBIx::HA::activeserver{$dbname});
+					return 0;
 				}
-			} else {
-				warn "$prefix in _readsharedfile: $dbname shared file has erroneous content, overwriting.\n";
-				_writesharedfile($dbname, $DBIx::HA::activeserver{$dbname});
-				return 0;
 			}
 		}
 	}
@@ -192,10 +174,6 @@ sub _isactivedb {
 	my $dsn = shift;
 	my $dbname = _getdbname ($dsn);
 	_readsharedfile($dbname);
-	if ($DATABASE::conf{$dbname}->{'end_of_stack'}) {
-		# we're not in the active db, because there is no active database, the end of stack is reached.
-		return 0;
-	}
 	if ($dsn eq $DATABASE::conf{$dbname}->{'active_db'}->[0]) {
 		warn "$prefix in _isactivedb: ".$dsn." is the active one \n" if (DBIx_HA_DEBUG > 2);
 		return 1;
@@ -229,14 +207,10 @@ sub _getnextdb {
 			$idxnextdb = 0;
 		} elsif ($idxnextdb > ($stackcount - 1)) {
 			warn "$prefix in _getnextdb: Reached end of db server stack for $dbname. Staying there.\n" if (DBIx_HA_DEBUG);
-			$DATABASE::conf{$dbname}->{'end_of_stack'} = 1;
-			_writesharedfile($dbname, -1);
-			return undef;
+			$idxnextdb = $stackcount - 1;
 		}
 		_writesharedfile($dbname, $idxnextdb);
 		warn "$prefix in _getnextdb: activated ".$DATABASE::conf{$dbname}->{'active_db'}->[0]." \n" if (DBIx_HA_DEBUG);
-	} elsif ($DATABASE::conf{$dbname}->{'end_of_stack'}) {
-		return undef;
 	} else {
 		warn "$prefix in _getnextdb: found different active db server, switching to ".$DATABASE::conf{$dbname}->{'active_db'}->[0]."\n" if (DBIx_HA_DEBUG);
 	}
@@ -263,7 +237,7 @@ sub _getApacheDBIidx {
 
 sub _reconnect {
 	my $currdsn = shift;
-	my $dbh = shift || undef;
+	our $dbh = shift || undef;
 	my $olddsn = $currdsn;	# old dsn to delete from Apache::DBI
 	my $conn_str;
 	my $selrow;
@@ -274,11 +248,6 @@ sub _reconnect {
 	if (! _isactivedb ($currdsn)) {	# wrong database server, use the active one
 		$currdsn = _getnextdb ($currdsn);
 	}
-	if (! $currdsn) {
-		warn "$prefix in _reconnect: No data source available, end of stack is reached. dbh is undefined. \n";
-		eval { $dbh->disconnect if ((defined $dbh) && (ref $dbh)); undef $dbh; };
-		return ($currdsn, undef);	# bad dbh and dsn
-	}
 
 	FINDDB: {
 	my $dbstackindex = 0;	# pointer to position in the stack
@@ -294,32 +263,24 @@ sub _reconnect {
 			$DATABASE::retries{$currdsn} = 0 if (! $DATABASE::retries{$currdsn});
 			for ($i=$DATABASE::retries{$currdsn}; $i < $DATABASE::conf{$dbname}->{'max_retries'}; $i++) {	# retry max_retries
 				$DATABASE::retries{$currdsn}++;
-				# now we're going to create a new good database handle and swap it with the old bad one
+				# now try to destroy, clear, undefine every instance and pointer of and to the $dbh
 				$newdbh = _connect_with_timeout (@$selrow);
 				if (defined $newdbh) {
-					# we managed to create a new database handle
-					if ((defined $dbh) && (ref $dbh)) {
-						# the old one still exists, so we're going to swap it and then destroy it
+					# all is good
+					if (defined $dbh) {
 						warn "$prefix in _reconnect: Pointing dbh to newdbh\n" if (DBIx_HA_DEBUG);
 						$dbh->swap_inner_handle($newdbh);
-						# wipe the old database handle (which in turn finishes all its children sth's)
-						# If we're using DBD::Sybase, make sure syb_flush_finish is off so we don't get remaining results
-						if ($dbh->{Driver}->{Name} eq 'Sybase') {
-							$newdbh->{syb_flush_finish} = 0;
-						}
-						eval { undef $newdbh; };
+						undef $newdbh;
 					} else {
-						# there was no previous active database handle, so that's easy
 						$dbh = $newdbh;
 					}
-					warn "$prefix Successfully reconnected to $currdsn\n";
-					$DATABASE::retries{$currdsn} = 0; # reset the retries counter
+					warn "$prefix in _reconnect: connected to $currdsn\n" if (DBIx_HA_DEBUG);
 					_writesharedfile($dbname, $dbstackindex);
 					# Do callback if it exists
 					if ( ref $DATABASE::conf{$dbname}->{'callback_function'}) {
 						&{$DATABASE::conf{$dbname}->{'callback_function'}}($dbh, $dbname);
 					}
-					return ($currdsn, $dbh); 
+					return ($currdsn); 
 				} #if
 				warn "$prefix in _reconnect: failed ".($i+1)." times to connect to $currdsn\n" if (DBIx_HA_DEBUG > 1);
 				select undef, undef, undef, 0.2; # wait a fraction of a second
@@ -328,11 +289,7 @@ sub _reconnect {
 			# Get another one, and try again, assuming we've not exhausted the stack!
 			$olddsn = $currdsn;			# remember the old one to delete it from Apache::DBI
 			$currdsn = _getnextdb ($currdsn);	# go to next dsn
-			if (! $currdsn) {			# we reached the end of the stack!
-				warn "$prefix *** ERROR: Exhausted DBI failover stack. Last DSN is: $olddsn \n";
-				return ($olddsn, undef);
-			}
-			warn "$prefix in _reconnect: dbstackindex: $dbstackindex; Trying another db server: $currdsn \n";
+			warn "$prefix in _reconnect: dbstackindex: $dbstackindex; Trying another db server: $currdsn \n" if (DBIx_HA_DEBUG);
 			goto FINDDB;
 		} #if
 		$dbstackindex++;
@@ -345,6 +302,7 @@ sub _reconnect {
 sub connect {
 	warn "$prefix Apache::DBI handlers are: \n" if (DBIx_HA_DEBUG > 1);
 	warn Dumper Apache::DBI::all_handlers() if (DBIx_HA_DEBUG > 1 && $loaded_Apache_DBI);
+
 	my $class = shift;
 	my $dbname = shift;
 	my ($dsn, $username, $auth, $attrs) = @{$DATABASE::conf{$dbname}->{'active_db'}};
@@ -357,12 +315,12 @@ sub connect {
 
 	# now we've got the right data source. Go ahead.
 	$DATABASE::retries{$dsn} = 0;	# initialize # of retries for the dsn
-	my $dbh = _connect_with_timeout($dsn, $username, $auth, $attrs);
+	our $dbh = _connect_with_timeout($dsn, $username, $auth, $attrs);
 	if (defined $dbh) {
 		warn "$prefix in connect: first try worked for $dsn\n" if (DBIx_HA_DEBUG);
 	} else {
-		warn "$prefix in connect: retrying connect of $dsn\n" if (DBIx_HA_DEBUG);
-		($dsn, $dbh) = _reconnect ($dsn, $dbh);
+		warn "$prefix in connect: retrying connect of $dsn\n" if (DBIx_HA_DEBUG > 1);
+		$dsn = _reconnect ($dsn);
 	}
 	return $dbh;
 }
@@ -371,7 +329,7 @@ sub _connect_with_timeout {
 	my ($dsn, $username, $auth, $attrs) = @_;
 	warn "$prefix in _connect_with_timeout: dsn: $dsn \n" if (DBIx_HA_DEBUG > 1);
 	my $res;
-	my $dbh;
+	our $dbh;
 	my $timeout = 0;
 	eval {
 		no strict;
@@ -387,7 +345,7 @@ sub _connect_with_timeout {
 	alarm(0);
 	if ($@ or $timeout) {	# there's a problem above
 		if ($timeout) {	# it's a timeout
-			warn "$prefix *** CONNECT TIMED OUT in $dsn";
+			warn "$prefix CONNECT TIMED OUT in $dsn";
 			eval { $dbh->disconnect };
 			$dbh = undef;
 		} else {	# problem in the connection
@@ -400,7 +358,6 @@ sub _connect_with_timeout {
 
 {
 package DBIx::HA::db;
-use strict;
 use constant DBIx_HA_DEBUG => DBIx::HA::DBIx_HA_DEBUG;
 use vars qw ( @ISA );
 @ISA = qw(DBI::db DBIx::HA);
@@ -408,16 +365,16 @@ our $prefix = "[$$] DBIx::HA:db:        ";
 
 # note that the DBI::db methods do not fail if the database connection is dead
 sub prepare {
-	my $dbh = shift;
+	our $dbh = shift;
 	my $sql = shift;
-	my $sth;
+	our $sth;
 	my $dsn = 'dbi:'.$dbh->{Driver}->{Name}.':'.$dbh->{Name};
 	warn "$prefix in prepare: dsn: $dsn ; sql: $sql \n" if (DBIx_HA_DEBUG > 1);
 	if (DBIx::HA::_isactivedb ($dsn)) {
-		warn "$prefix Statement handle being prepared while existing statement handle still open!\n\tdbh:\t\t$dsn\n\tprevious statement:\t".$dbh->{Statement}."\n\tcurrent statement:\t$sql\n\tACTIVE KIDS: ".$dbh->{ActiveKids}."\n" if ($dbh->{ActiveKids});
+		warn "Statement handle being prepared while existing statement handle still open!\n\tdbh:\t\t$dsn\n\tprevious statement:\t".$dbh->{Statement}."\n\tcurrent statement:\t$sql\nACTIVE KIDS: ".$dbh->{ActiveKids}."\n" if ($dbh->{ActiveKids});
 	} else {
 		my $dbname = DBIx::HA::_getdbname($dsn);
-		($dsn, $dbh) = DBIx::HA::_reconnect ($dsn, $dbh);
+		$dsn = DBIx::HA::_reconnect ($dsn, $dbh);
 		if (! defined $dbh) { # we couldn't connect at all
 			warn "$prefix in prepare: couldn't prepare sql: $sql\n";
 			return undef;
@@ -431,7 +388,6 @@ sub prepare {
 
 {
 package DBIx::HA::st;
-use strict;
 use constant DBIx_HA_DEBUG => DBIx::HA::DBIx_HA_DEBUG;
 use Sys::SigAction qw( set_sig_handler );
 use vars qw ( @ISA $prefix );
@@ -439,63 +395,62 @@ use vars qw ( @ISA $prefix );
 our $prefix = "[$$] DBIx::HA:st:        "; 
 
 sub execute {
-	my $sth = shift;
-	my $dbh = $sth->{Database};
+	our $sth = shift;
+	our $dbh = $sth->{Database};
 	my $sql = $dbh->{Statement};
 	my $dsn = 'dbi:'.$dbh->{Driver}->{Name}.':'.$dbh->{Name};
-	my $dbname = DBIx::HA::_getdbname($dsn);
 	my $res;
-	my $to;	# did we trip a timeout on the execution?
-	my $orig_error_code;
-	my $orig_error_string;
-	my $max_executions = $DATABASE::conf{$dbname}->{'max_retries'} * scalar(@{$DATABASE::conf{$dbname}->{'db_stack'}});
 
-	warn "=================\n" if (DBIx_HA_DEBUG > 1);
 	warn "$prefix in execute: dsn: $dsn ; sql: $sql \n" if (DBIx_HA_DEBUG > 1);
 	if (DBIx::HA::_isactivedb ($dsn)) {
-		($res, $to) = &_execute_with_timeout ($dsn, $sth);
-		$orig_error_code = $DBI::err;
-		$orig_error_string = $DBI::errstr;
-		if ($to || ((! defined $res) && &{$DATABASE::conf{$dbname}->{'failtest_function'}}($orig_error_code, $orig_error_string))) {
-			# It was a timeout error or a critical network library error (connection in a bad state)
-			warn "$prefix in execute: timeout error or network lib error, reexecuting: $sql ; dsn: $dsn \n" if (DBIx_HA_DEBUG);
-			for (my $count_execs = 0; $count_execs < $max_executions; $count_execs++) {
+		$res = &_execute_with_timeout ($dsn, $sth);
+		if (! defined $res) {
+			eval {
+				$sth->finish;
+			};
+			# first try a simple statement. If it fails, then we should reconnect.
+			my $res2 = $dbh->do('select 1');
+
+			if (! defined $res2) {
+				# Ooops. Even a simple statement fails. It's time to reconnect and reexecute
+				warn "$prefix in execute: execution failed, attempting reexecution. statement: $sql ; dsn: $dsn \n" if (DBIx_HA_DEBUG);
 				($dsn, $sth, $res) = _reexecute ($dsn, $sql, $sth);
-				last if (defined $res); # reexecution worked (or failed hard with -666)
-			} 
-		} elsif (! defined $res) {
-			# We got an error code from the server upon statement execution.
-			# We will let the client decide what to do and let it be.
-			warn "$prefix *** ERROR: $orig_error_code; $orig_error_string \n" if (DBIx_HA_DEBUG);
-			warn "$prefix in execute: bad sql: $sql ; dsn: $dsn \n" if (DBIx_HA_DEBUG);
+			} else {
+				# here we have a problem: the original statement didn't execute but "select 1" did
+				# 2 choices:
+				#	1- the sql code was invalid
+				#	2- the db server was so busy that it timed out on the original statement only
+				# we will wait a short time and then try again just once
+				warn "$prefix in execute: server busy or bad sql: $sql ; dsn: $dsn \n" if (DBIx_HA_DEBUG);
+				select undef, undef, undef, 0.2;	# wait 200ms
+				my $newsth = $dbh->prepare($sql);
+				$res = _execute_with_timeout ($dsn, $newsth);
+				$sth->swap_inner_handle($newsth);
+				undef $newsth;
+			}
 		}
 	} else { # current db is not active
-		for (my $count_execs = 0; $count_execs < $max_executions; $count_execs++) {
-			($dsn, $sth, $res) = _reexecute ($dsn, $sql, $sth);
-			last if (defined $res); # reexecution worked (or failed hard with -666)
-		}
+		eval { $sth->finish; };
+		($dsn, $sth, $res) = _reexecute ($dsn, $sql, $sth);
 	}
-	if (! defined $res) { # Execution failed
-		warn "$prefix in execute: result is undefined, statement execution failed! statement: $sql ; dsn: $dsn \n" if (DBIx_HA_DEBUG);
-		warn "+++++++++++++++++\n" if (DBIx_HA_DEBUG > 1);
+	if (! defined $res) { # what the hell?
+		warn "$prefix in execute: result is undefined, statement execution failed! statement: $sql ; dsn: $dsn \n";
 		return undef;
 	}
 	if ($res == -666) { # the devil killed you! We couldn't connect to the db!
-		warn "$prefix in execute: statement couldn't be executed because connect failed abnormally. statement: $sql ; dsn: $dsn \n";
-		warn "+++++++++++++++++\n" if (DBIx_HA_DEBUG > 1);
+		warn "$prefix in execute: statement couldn't be executed because connect failed after many retries! statement: $sql ; dsn: $dsn \n" if (DBIx_HA_DEBUG);
 		return undef;
 	}
 	warn "$prefix in execute: statement executed successfully! statement: $sql ; dsn: $dsn \n" if (DBIx_HA_DEBUG);
 	warn "$prefix in execute: res: $res ; errstr: $DBI::errstr \n" if (DBIx_HA_DEBUG);
 	$DATABASE::retries{$dsn} = 0;	# flush the retries to this dsn, since executing worked
 	undef $@; # don't make an upstream eval die because of what happened here, since we're fine now
-	warn "+++++++++++++++++\n" if (DBIx_HA_DEBUG > 1);
 	return $res;
 }
 
 sub _execute_with_timeout {
 	my $dsn = shift;
-	my $sth = shift;
+	our $sth = shift;
 	my $sql = $sth->{Statement};
 	warn "$prefix in _execute_with_timeout: dsn: $dsn ; sql : $sql \n" if (DBIx_HA_DEBUG > 1);
 	my $res;
@@ -514,51 +469,49 @@ sub _execute_with_timeout {
 	alarm(0);
 	if ($@ or $timeout) {	# there's a problem above
 		if ($timeout) {	# it's a timeout
-			warn "$prefix *** EXECUTION TIMED OUT in $dsn ; SQL: $sql";
+			warn "$prefix EXECUTION TIMED OUT in $dsn ; SQL: $sql";
+			eval { $sth->finish; };
+			$dbh = undef;
 		} else {	# problem in the execution
 			warn "$prefix Error in DBI::execute: $@\n" if $@;
 		}
-		eval { $sth->finish; };
-		$sth = undef;
-		$res = undef;
 	}
-	return ($res, $timeout);
+	return $res;
 }
 
 sub _reexecute {
 	# reexecute the statement in the following way:
 	# reconnect with a new dbh
-	# redo prepare and execute until it works
+	# redo prepare and execute
+	# only do it once, since if the connection works and the execution doesn't,
+	# it means that the statement sql is wrong and should not be retried.
 	my $dsn = shift;
 	my $sql = shift;
-	my $sth = shift || undef;
-	my $dbh = undef;
+	our $sth = shift || undef;
+	our $dbh = undef;
 	my $newsth;
 	my $res;
-	my $to;
 
 	warn "$prefix in _reexecute: dsn: $dsn \n" if (DBIx_HA_DEBUG > 1);
 	warn "$prefix Reexecuting statement: $sql" if (DBIx_HA_DEBUG > 1);
+	my $dbname = DBIx::HA::_getdbname($dsn);
 	if (defined $sth) {
 		$dbh = $sth->{Database};
 	}
-	($dsn, $dbh) = DBIx::HA::_reconnect ($dsn, $dbh);
-	if (! defined $dbh) { return ($dsn, $sth, -666); } # we couldn't connect at all
+	$dsn = DBIx::HA::_reconnect ($dsn, $dbh);
+	if (! defined $dbh) { return ($dsn, -666); } # we couldn't connect at all
 	$newsth = $dbh->prepare($sql);
-	($res, $to) = &_execute_with_timeout ($dsn, $newsth);
-	if (! $res) {	# execute_with_timeout failed
-		warn "$prefix in _reexecute: reexecuting failed. dsn: $dsn  ; statement: $sql\n" if (DBIx_HA_DEBUG);
-		eval { $sth->finish; };
-		return ($dsn, $sth, undef);
-	}
+	$res = &_execute_with_timeout ($dsn, $newsth);
 	if (defined $sth) {
-		$sth->swap_inner_handle($newsth, 1);	# allow reparenting of the statement handle
-		eval { $newsth->finish; };
+		$sth->swap_inner_handle($newsth);
 		undef $newsth;
 	} else {
 		$sth = $newsth;
 	}
-	return ($dsn, $sth, $res);
+	return ($dsn, $sth, $res) if (defined $res);
+	warn "$prefix in _reexecute: reexecuting failed. dsn: $dsn  ; statement: $sql\n" if (DBIx_HA_DEBUG);
+
+	return ($dsn, $sth, undef);
 }
 
 } # end package DBIx::HA::st
@@ -597,7 +550,6 @@ DBIx::HA - High Availability package for DBI
     connecttimeout  => 1,
     executetimeout  => 8,
     callback_function => \&MyCallbackFunction,
-	failtest_function   => \&DBIx::HA::FTF_SybaseASE,
     };
 
  DBIx::HA->initialize();
@@ -631,15 +583,13 @@ back online before it is put back in service.
 =item timeouts
 
 Database calls are wrapped in user-configurable timeouts. Connect and execute
-timeouts are handled independently. As of version 0.62, timeouts are 
+timeouts are handled independently. As of version 0.62, timeouts are now
 handled through Sys::SigAction for consistent signal handling behavior across
 Perl versions.
 
 =item configurable retries
 
 Queries can be retried n times before a datasource is considered failed.
-Starting with version 0.95, the retry counter is reset whenever a reconnect
-works.
 
 =item callback function
 
@@ -685,11 +635,11 @@ first entry is the main dsn that will be used at start.
 =item max_retries REQUIRED
 
 max_retries takes an integer > 0 as value. It determines the number of times
-a datasource will be consecutively retried upon failure. It B<is reset> upon
-success of a retry. This is a change in behavior starting in version 0.95.
-For example, if max_retries is 3, if datasource #1 can't be reached three
-times in a row then I<_reconnect()> will reset the number of tries and go to
-datasource #2 if available.
+a datasource will be retried upon failure. It is B<NOT> reset upon success
+unless the next datasource is reached. For example, if datasource #1 was
+already retried once before, and max_retries is 3, if datasource #1 can't be
+reached twice in a row then I<_reconnect()> will reset the number of tries
+and go to datasource #2 if available.
 
 =item connectoninit ( DEFAULT: 0 )
 
@@ -733,35 +683,13 @@ timeout for connecting to a datasource, in seconds.
 
 =item executetimeout ( DEFAULT: 10 )
 
-timeout for execution of a statement, in seconds. If the timeout is triggered,
-the database handle is deleted and a new connect is tried. If the connect
-succeeds, we assume that the problem is with a runaway SQL statement or bad
-indexing. If the connect fails, then we fail over.
+timeout for preparation or execution of a statement, in seconds.
 
 =item callback_function ( DEFAULT: I<none> )
 
 reference to a function to call whenever the datasource is changed due to a
 failover. See the TIPS sections for a usage example.
 
-=item failtest_function ( DEFAULT: sub{0} )
-
-reference to a function to call to test if a DBI error is a candidate for
-failover or not.
-Input is ($ErrorID, $ErrorString)
-Output is boolean: If true, then we'll consider the error a critical
-condition, ok to failover. If false, then DBIx::HA will not act on it
-and pass it straight through to the client.
-
-This Fail Test Function (FTF) function is extremely important for the proper
-functioning of DBIx::HA. You must be careful  in defining it
-precisely, based on the database engine that you are using. As a
-convenience to the user, DBIx::HA defines some sample functions that
-you can use. At this time, only the following samples are defined:
-
- DBIx::HA::FTF_SybaseASE
-
-Example use is in the synopsis
-
 =back
 
 =head1 USER METHODS
@@ -858,19 +786,11 @@ directory in your Apache installation. Assuming that you are using
 B<failoverlevel 'application'>, all processes will switch to the datasource you
 define in that file as soon as they are ready to prepare or execute a statement.
 
-Another trick is to set the value in the shared file to -1. This will tell the
-module that we've reached the end of the stack and no connection should be
-attempted, effectively blocking all database calls.
-
-Conversely, if the shared file does contain -1 because all DSNs in the stack
-have failed, you can reset it to whatever DSN entry you want without having to
-bounce Apache.
-
 =back
 
 =head1 DEPENDENCIES
 
-This modules requires Perl >= 5.6.0, DBI >= 1.49  and Sys::SigAction.
+This modules requires Perl >= 5.6.0, DBI >= 1.44  and Sys::SigAction.
 Apache::DBI is recommended when using mod_perl.
 If using Apache::DBI, version 0.89 or above is required.
 Always load Apache::DBI and Apache before DBIx::HA if you want DBIx::HA to know
@@ -895,7 +815,7 @@ Henri Asseily <henri@asseily.com>
 
 =head1 COPYRIGHT
 
-Copyright (c) 2003-2006 Henri Asseily <henri@asseily.com>.
+Copyright (c) 2003-2004 Henri Asseily <henri@asseily.com>.
 All rights reserved. This program is free software; you can redistribute
 it and/or modify it under the same terms as Perl itself.
 
